from unittest.mock import patch

import pytest
from id_broker_handlers import create_work_group_if_missing

from idbroker.id_broker_client import IDBrokerNotFoundError, IDBrokerWorkGroup as WorkgroupClient
from ucsschool_id_connector.models import SchoolUserRole


@pytest.mark.asyncio
async def test_handle_attr_context(
    user_dispatcher, get_listener_user_add_modify_object, user_dn, workgroup_dn, school_class_dn
):
    listener_obj = {
        "school": ["DEMOSCHOOL"],
        "schools": ["DEMOSCHOOL", "DEMOSCHOOL-2"],
        "groups": [workgroup_dn, school_class_dn],
    }
    obj = get_listener_user_add_modify_object(dn=user_dn, object=listener_obj)
    context = await user_dispatcher._handle_attr_context(obj=obj)
    assert context["DEMOSCHOOL"]["workgroups"] == {"wg1"}
    assert context["DEMOSCHOOL"]["classes"] == {"1a"}
    assert context["DEMOSCHOOL"]["roles"] == {SchoolUserRole.teacher}


@pytest.mark.asyncio
async def test__empty_groups(
    user_dispatcher,
):
    context_text = {"DEMOSCHOOL": {"classes": ["my-test-class"], "workgroups": ["my-test-workgroup"]}}
    context = user_dispatcher._empty_groups(context_text)
    assert context["DEMOSCHOOL"]["classes"] == []
    assert context["DEMOSCHOOL"]["workgroups"] == []


class MockValue(object):
    def __init__(self, value):
        self.value = value


class MockObject(object):
    def __init__(self, entry_uuid, description):
        self.entryUUID = MockValue(entry_uuid)
        self.description = MockValue(description)


class MockLDAPAccess:
    called_with_filter_s = None
    called_with_attributes = None
    entries_in_back_end = None

    async def search(self, filter_s, attributes):
        MockLDAPAccess.called_with_filter_s = filter_s
        MockLDAPAccess.called_with_attributes = attributes
        for dn, attrs in MockLDAPAccess.entries_in_back_end:
            if filter_s == dn:
                return [attrs]
        return []


@pytest.mark.asyncio
async def test_create_workgroup_if_missing(school_authority_conf, user_dispatcher):
    wg_name = "wg1"
    mock_ldap_access = MockLDAPAccess()
    MockLDAPAccess.entries_in_back_end = [("", "")]
    logger = user_dispatcher.logger
    client = WorkgroupClient(school_authority=school_authority_conf, plugin_name="id_broker")
    with pytest.raises(IDBrokerNotFoundError):
        await create_work_group_if_missing(
            work_group=wg_name,
            ou="TEST",
            id_broker_work_group=client,
            ldap_access=mock_ldap_access,
            logger=logger,
        )

    MockLDAPAccess.entries_in_back_end = [
        (
            "(&(objectClass=univentionGroup)(cn=TEST-wg1))",
            MockObject(entry_uuid="test-entry-uuid", description="description"),
        )
    ]

    async def _exists(self, **kwargs):
        return True

    called_create = False

    async def _create(self, **kwargs):
        nonlocal called_create
        called_create = True

    with patch("idbroker.id_broker_client.ProvisioningAPIClient._exists", _exists):
        await create_work_group_if_missing(
            work_group=wg_name,
            ou="TEST",
            id_broker_work_group=client,
            ldap_access=mock_ldap_access,
            logger=logger,
        )
        assert called_create is False

    async def _exists(self, **kwargs):
        return False

    with patch("idbroker.id_broker_client.ProvisioningAPIClient._exists", _exists):
        with patch("idbroker.id_broker_client.ProvisioningAPIClient._create", _create):
            await create_work_group_if_missing(
                work_group=wg_name,
                ou="TEST",
                id_broker_work_group=client,
                ldap_access=mock_ldap_access,
                logger=logger,
            )
        assert called_create is True
